#define outputFile "!output.txt"
#define debugFile "!debug.txt"
#define infoFile "!info.txt"
#define programFile "program.txt"
#define inputFile "input.txt"

#include <fstream>
#include <string.h>
#include <vector>
#include <cstdlib>

using namespace std;

bool fexists(const char* filename) {
  ifstream ifile(filename);
  return ifile;
}

vector<string> s, r;
int maxTime = 1000, maxProgramLength = 100, maxSearchLength = 10, maxReplacementLength = 10;

int KMP(string S, string K)
{
    vector<int> T(K.size() + 1, -1);

	for (int i = 1; i <= K.size(); i++)
	{
		int pos = T[i - 1];
		while (pos != -1 && K[pos] != K[i - 1])
            pos = T[pos];
		T[i] = pos + 1;
	}

	int sp = 0, kp = 0;
	while(sp < S.size())
	{
		while (kp != -1 && (kp == K.size() || K[kp] != S[sp]))
            kp = T[kp];
		kp++;
		sp++;
		if (kp == K.size())
            return sp - K.size();
	}

	return -1;
}

int automaton(vector<string> s, vector<string> r, string in, int limit, string &out, FILE* fdebug)
{
    out = in;
    int time = 0;
    while(true)
    {
        if (fdebug != NULL)
            fprintf(fdebug, "%s\n", out.c_str());
        bool replaced = false;
        for (int i = 0; i < s.size(); i++)
        {
            int pos = KMP(out, s[i]);
            if (pos >= 0)
            {
                time++;
                if (time > limit)
                    return -1;
                out = out.substr(0, pos) + r[i] + out.substr(pos + s[i].size());
                replaced = true;
                break;
            }
        }
        if (! replaced)
            return time;
    }
}

int main(int argc, char* argv[])
{
    if (fexists(outputFile))
        remove(outputFile);
    if (fexists(debugFile))
        remove(debugFile);
    FILE* finfo = fopen(infoFile, "w");

    if (! fexists(programFile))
    {
        fprintf(finfo, "   program.txt\n");
        fclose(finfo);
        return 0;
    }
    if (! fexists(inputFile))
    {
        fprintf(finfo, "   input.txt\n");
        fclose(finfo);
        return 0;
    }

    if (argc >= 2)
    {
        maxTime = atoi(argv[1]);
        if (maxTime < 1 || maxTime > 10000)
        {
            fprintf(finfo, "         1  10000\n");
            fclose(finfo);
            return 0;
        }
    }

    if (argc >= 3)
    {
        maxProgramLength = atoi(argv[2]);
        if (maxProgramLength < 1 || maxProgramLength > 1000)
        {
            fprintf(finfo, "         1  1000\n");
            fclose(finfo);
            return 0;
        }
    }

    if (argc >= 4)
    {
        maxSearchLength = atoi(argv[3]);
        if (maxSearchLength < 1 || maxSearchLength > 100)
        {
            fprintf(finfo, "          1  100\n");
            fclose(finfo);
            return 0;
        }
    }

    if (argc >= 5)
    {
        maxReplacementLength = atoi(argv[4]);
        if (maxReplacementLength < 1 || maxReplacementLength > 100)
        {
            fprintf(finfo, "          1  100\n");
            fclose(finfo);
            return 0;
        }
    }

    FILE* fprogram = fopen(programFile, "r");
    FILE* finput = fopen(inputFile, "r");

    char c;
    int mode = 0; // Looking for a new rule
    int emptyLines = 0, programLength = 0;
    string searchString = "", replacementString = "";
    while (true)
    {
        int readCount = fscanf(fprogram, "%c", &c);
        if (readCount <= 0 || c == '\r' || c == '\n')
        {
            if (mode != 0)
            {
                s.push_back(searchString);
                r.push_back(replacementString);
                searchString = "";
                replacementString = "";
                mode = 0; // Looking for a new rule
            } else if (readCount > 0 && c == '\n')
            {
                emptyLines++;
                if (emptyLines > 1)
                {
                    fprintf(finfo, "ҳ       ");
                    fclose(finfo);
                    fclose(fprogram);
                    fclose(finput);
                    return 0;
                }
            }
        } else if (c == ' ')
        {
            if (mode != 1)
            {
                fprintf(finfo, "ҳ      ");
                fclose(finfo);
                fclose(fprogram);
                fclose(finput);
                return 0;
            }
            mode = 2; // Expecting the replacement string
        } else if (c == '+' || c >= '0' && c <= '9' || c >= 'a' && c <= 'z' || c >= 'A' && c <= 'Z')
        {
            if (mode == 0)
            {
                searchString += c;
                mode = 1; // Reading search string
                emptyLines = 0;
                programLength++;
                if (programLength > maxProgramLength)
                {
                    fprintf(finfo, "    ");
                    fclose(finfo);
                    fclose(fprogram);
                    fclose(finput);
                    return 0;
                }
            } else if (mode == 1)
            {
                searchString += c;
                if (searchString.size() > maxSearchLength)
                {
                    fprintf(finfo, "     ");
                    fclose(finfo);
                    fclose(fprogram);
                    fclose(finput);
                    return 0;
                }
            } else if (mode == 2)
            {
                replacementString += c;
                mode = 3; // Reading replacement string
            } else if (mode == 3)
            {
                replacementString += c;
                if (replacementString.size() > maxReplacementLength)
                {
                    fprintf(finfo, "     ");
                    fclose(finfo);
                    fclose(fprogram);
                    fclose(finput);
                    return 0;
                }
            }
        } else
        {
            fprintf(finfo, "    : %c", c);
            fclose(finfo);
            fclose(fprogram);
            fclose(finput);
            return 0;
        }
        if (readCount <= 0)
            break;
    }

    string inputString = "";
    while (fscanf(finput, "%c", &c) > 0)
    {
        if (c == '\r' || c == '\n')
            break;
        else if (c == '+' || c >= '0' && c <= '9' || c >= 'a' && c <= 'z' || c >= 'A' && c <= 'Z')
            inputString += c;
        else
        {
            fprintf(finfo, "    : %c", c);
            fclose(finfo);
            fclose(fprogram);
            fclose(finput);
            return 0;
        }
    }

    FILE* fdebug = fopen(debugFile, "w");

    string outString;
    int time = automaton(s, r, inputString, maxTime, outString, fdebug);
    if (time == -1)
    {
        fprintf(finfo, "    ");
        fclose(finfo);
        fclose(fprogram);
        fclose(finput);
        fclose(fdebug);
        return 0;
    }

    FILE* foutput = fopen(outputFile, "w");
    fprintf(foutput, "%s\n", outString.c_str());

    int searchLength = 0, replacementLength = 0;
    for (int i = 0; i < programLength; i++)
    {
        if (s[i].size() > searchLength)
            searchLength = s[i].size();
        if (r[i].size() > replacementLength)
            replacementLength = r[i].size();
    }
    fprintf(finfo, "  \nʳ : %d\n : %d\n  : %d (), %d ()\n", time, programLength, searchLength, replacementLength);

    fclose(finfo);
    fclose(fprogram);
    fclose(finput);
    fclose(fdebug);
    fclose(foutput);
    return 0;
}
